#include "functions.h"
#include "ping.h"

#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <math.h>

void generate_graph(pip_addr pip);

int main(int argc, char *argv[]) {

    // Verificacion del numero de parametros ingresados
    if (argc != 4) {
        if ( (argc == 2) && (strcmp(argv[1], "--help") == 0) ) {
            printf("Sintaxis:\n");
            printf("\trtdelay <host_name> <n> <max_size>\n\n");
            printf("Parametros:\n");
            printf("<host_name>     Direccion IP o URL del host destino\n");
            printf("<n>             Numero de mediciones a realizar\n");
            printf("<max_size>      Tamano maximo de los paquetes\n");
        }
        else {
            printf("Usage: rtdelay <host_name> <n> <max_size>\n");
            printf("Try 'rtdelay --help' for more information\n");
        }
        exit(1);
    }

    FILE *pf;
    char *cmd;
    char *line;
    pip_addr start;
    pip_addr ip_host;
    int ttl;

    cmd = (char *)malloc(sizeof(char)*BUFFER_SIZE);
    line = (char *)malloc(sizeof(char)*BUFFER_SIZE);

    // Determinar max_size para poder realizar mediciones con
    // tamanos uniformemente distribuidos.
    int size_conv = ((atoi(argv[3]) - 10) / 20) * 20;
    if (size_conv < 20) {
        printf("<max_size> insuficiente. Debe ingresar un valor igual o superior a 30.\n");
        exit(1);
    }

    // Determinar numero de saltos para llegar hasta el
    // el <host_name> especificado por el usuario.
    ttl = 0;
    ip_host = CreateIP();
    start = ip_host;
    do {
        strcpy(line, "");
        sprintf(cmd, "ping -t %d -c 1 %s | grep From | tr -d '():' | awk '{i = 0; while ($(++i) != \"Time\"); print $(i-2)}'", ++ttl, argv[1]);

        if ((pf = popen(cmd, "r")) == NULL) {
            perror("popen");
            exit(1);
        }

        fgets(line, BUFFER_SIZE, pf);
        if (strcmp(line, "") != 0) {
            line = remove_nl(line);
            strcpy(ip_host->address, line);
            ip_host->hop = ttl;
            ip_host->size_step = size_conv / 20;
            ip_host->measures = atoi(argv[2]);
            ip_host->next = CreateIP();
            ip_host = ip_host->next;
        }
    } while(strcmp(line, "") != 0);

    strcpy(line, "");
    sprintf(cmd, "ping -c 1 %s | grep icmp | tr -d '():' | awk '{i = 0; while ($(++i) != \"icmp_req=1\"); print $(i-1)}'", argv[1]);

    if ((pf = popen(cmd, "r")) == NULL) {
        perror("popen");
        exit(1);
    }

    fgets(line, BUFFER_SIZE, pf);
    if (strcmp(line, "") != 0) {
        line = remove_nl(line);
        strcpy(ip_host->address, line);
        ip_host->hop = ttl;
        ip_host->size_step = size_conv / 20;
        ip_host->measures = atoi(argv[2]);
    }
    
    printf("Saltos hasta %s: %d\n", argv[1], ttl);

    // Realizacion de los ping a cada router hasta el
    // host destino
    pthread_t thread[ttl];
    pip_addr pip;
    int i;

    printf("Pinging... ");

    for ( i = 0, pip = start; pip != NULL; pip = pip->next, i++)
        pthread_create(&thread[i], NULL, ping_task, pip);

    for ( i = 0; i < ttl; i++)
        pthread_join(thread[i], NULL);

    printf("Done\n");

    // Generacion de los graficos
    generate_graph(start);

    free(cmd);
    free(line);
    pclose(pf);

    exit(0);
}


void generate_graph(pip_addr pip) {
    id_t pid;
    int to_matlab[2];
    int from_matlab[2];
    int i, status;
    float slope;
    char *buffer;
    FILE *sd;
    FILE *fd;
    pip_addr pip_tmp;

    buffer = (char *)malloc(sizeof(char)*BUFFER_SIZE);

    if (pipe(to_matlab) < 0) {
        perror("pipe");
        exit(1);
    }

    if (pipe(from_matlab) < 0) {
        perror("pipe");
        exit(1);
    }

    if ((pid = fork()) < 0) {
        perror("fork");
        exit(1);
    }

    // Proceso hijo que ejecuta Matlab
    if (pid == 0) {
        int junk;

        dup2(to_matlab[0], 0);
        dup2(from_matlab[1], 1);
        close(to_matlab[1]);
        close(from_matlab[0]);
        junk = open("/dev/null", O_WRONLY);
        dup2(junk, 2);
        execlp("matlab", "matlab", "-nodesktop", NULL);
        perror("exec");
        exit(-1);
    }

    close(to_matlab[0]);
    close(from_matlab[1]);
    sd = fdopen(to_matlab[1], "w");
    fd = fdopen(from_matlab[0], "r");

    fprintf(sd, "size = [");
    for ( i = 0; i < 20; i++)
        fprintf(sd, "%d; ", (10 + i*pip->size_step));
    fprintf(sd, "];\n");
    fflush(sd);

    for ( pip_tmp = pip; pip_tmp != NULL; pip_tmp = pip_tmp->next) {
        fprintf(sd, "rtt%d = [", pip_tmp->hop);
        for ( i = 0; i < 20; i++)
            fprintf(sd, "%f ;", pip_tmp->rtt[i]);
        fprintf(sd, "];\n");
    }
    fflush(sd);

    // Primer grafico de retardo de cada router
    // hasta el host destino.
    fprintf(sd, "rtt = [");
    for ( pip_tmp = pip; pip_tmp != NULL; pip_tmp = pip_tmp->next)
        fprintf(sd, "rtt%d ", pip_tmp->hop);
    fprintf(sd, "];\n");
    fflush(sd);

    fprintf(sd, "h = figure;\n");
    fprintf(sd, "plot(size, rtt, '-s', 'LineWidth', 1.2);\n");

    fprintf(sd, "xlabel('Tamano de paquete [bytes]');\n");
    fprintf(sd, "ylabel('Retardo [ms]');\n");
    fprintf(sd, "title('Grafico de retardo de cada router hasta host destino');\n");
    fprintf(sd, "grid;\n");
    fprintf(sd, "xlim([10 %d]);\n", (10 + 20*pip->size_step));
    fprintf(sd, "legend(");
    for ( pip_tmp = pip; pip_tmp != NULL; pip_tmp = pip_tmp->next) {
        fprintf(sd, "'Salto #%d'", pip_tmp->hop);
        if (pip_tmp->next != NULL)
            fprintf(sd, ", ");
    }
    fprintf(sd, ",'Location','NorthWest');\n");
    fprintf(sd, "print(h, '-dpng', 'Figure1.png')\n");
    fflush(sd);

    // Segundo grafico de retardo estimado de cada enlace
    // hasta el host destino.
    for ( pip_tmp = pip; pip_tmp->next != NULL; pip_tmp = pip_tmp->next) {
        if (pip_tmp->measures_success) {
            fprintf(sd, "rtt_link%d = [", pip_tmp->hop);
            for ( i = 0; i < 20; i++)
                fprintf(sd, "%f ;", (pip_tmp->next->rtt[i] - pip_tmp->rtt[i]));
            fprintf(sd, "];\n");
        }
    }
    fflush(sd);

    fprintf(sd, "rtt_link = [");
    for ( pip_tmp = pip; pip_tmp->next != NULL; pip_tmp = pip_tmp->next)
        if (pip_tmp->measures_success)
            fprintf(sd, "rtt_link%d ", pip_tmp->hop);
    fprintf(sd, "];\n");
    fflush(sd);

    fprintf(sd, "h = figure;\n");
    fprintf(sd, "plot(size, rtt_link, '-s', 'LineWidth', 1.2);\n");

    fprintf(sd, "xlabel('Tamano de paquete [bytes]');\n");
    fprintf(sd, "ylabel('Retardo [ms]');\n");
    fprintf(sd, "title('Grafico de retardo estimado para cada enlace hasta host destino');\n");
    fprintf(sd, "grid;\n");
    fprintf(sd, "xlim([10 %d]);\n", (10 + 20*pip->size_step));
    fprintf(sd, "legend(");
    for ( pip_tmp = pip; pip_tmp->next != NULL; pip_tmp = pip_tmp->next) {
        if (pip_tmp->measures_success) {
            fprintf(sd, "'Enlace #%d'", pip_tmp->hop);
            if (pip_tmp->next->next != NULL)
                fprintf(sd, ", ");
        }
    }
    fprintf(sd, ",'Location','NorthWest');\n");
    fprintf(sd, "print(h, '-dpng', 'Figure2.png')\n");
    fflush(sd);

    // Obtencion de la pendiente del grafico de retardo estimado de cada
    // enlace mediante regresion lineal para la estimacion de la tasa del
    // enlace.
    i = 0;
    while ( i++ < 10 )
        fgets(buffer, BUFFER_SIZE, fd);

    for ( pip_tmp = pip; pip_tmp->next != NULL; pip_tmp = pip_tmp->next) {
        if (pip_tmp->measures_success) {
            fprintf(sd, "slope = polyfit(size, rtt_link%d,1);\n", pip_tmp->hop);
            fprintf(sd, "slope(1)\n");
            fflush(sd);

            fgets(buffer, BUFFER_SIZE, fd);
            fgets(buffer, BUFFER_SIZE, fd);
            fgets(buffer, BUFFER_SIZE, fd);
            fgets(buffer, BUFFER_SIZE, fd);

            slope = atof(remove_nl(&buffer[3]));
            printf("Tasa del enlace %d: %.3f Mbps\n", pip_tmp->hop, 8*(1/(fabs(slope)))/1000 );
            fgets(buffer, BUFFER_SIZE, fd);
        }
    }

    sleep(20);
    fprintf(sd, "clear all;\n");
    fprintf(sd, "\n exit\n");
    fflush(sd);

    waitpid(pid, &status, 0);
    free(buffer);
    fclose(sd);

    exit(0);
}
